#include <stdlib.h>
#include <string.h>

struct circular_buffer_type
{
  char *buf ;
  int buf_size ;
  int current_index ;  /* positioned 1 ahead of last byte written */
  int current_length ;
} ;

struct circular_buffer_type *circular_buffer_alloc()
{
  struct circular_buffer_type *result ;

  result=(struct circular_buffer_type *)malloc(sizeof(struct circular_buffer_type)) ;
  if(result!=NULL)
    {
      result->buf=NULL ;
      result->buf_size=0 ;
      result->current_index=0 ;
      result->current_length=0 ;
    }

  return(result) ;
}

char cb_nice_char(char x)
{
  if((x>=' ')&&(x<=0x7F))
    return(x) ;
  else
    return('?') ;
}

void circular_buffer_free(struct circular_buffer_type *cb)
{
  if(cb!=NULL)
    {
      if(cb->buf!=NULL)
        free(cb->buf) ;
      free(cb) ;
    }
}

int circular_buffer_realloc(struct circular_buffer_type *cb, int min, int max)
{
  int error_result, size, wrap ;
  char *new_buf ;

  error_result=0 ;
  size=cb->buf_size ;
  if(size<1)
    size=1 ;
  for(;size<min;)
    size*=2 ;
  if(max>0)
    if(size>max)
      size=max ;
  if(size<cb->current_length)
    size=cb->current_length ;

  /* if no need to transfer old data ... */
  if(cb->buf!=NULL)
    if(cb->current_length<1)
      {
        free(cb->buf) ;
        cb->buf=NULL ;
      }

  new_buf=(char *)malloc(size) ;
  if(new_buf==NULL)
    error_result= -1 ;
  else
    {
      /* transfer existing buf values */
      if(cb->current_length>0)
        {
          wrap=cb->current_length-cb->current_index ;
          if(wrap>0)
            {
              memcpy(new_buf,&(cb->buf[cb->buf_size-wrap]),wrap) ;
              memcpy(&(new_buf[wrap]),cb->buf,cb->current_index) ;
            }
          else
            memcpy(new_buf,&(cb->buf[cb->current_index-cb->current_length]),cb->current_length) ;
        }
      cb->current_index=cb->current_length ;
      cb->buf_size=size ;
      if(cb->buf!=NULL)
        free(cb->buf) ;
      cb->buf=new_buf ;
    }
  return(error_result) ;
}

int circular_buffer_write(struct circular_buffer_type *cb, char *b, int n)
{
  int error_result, space ;

  error_result=0 ;
  if(cb->buf_size<cb->current_length+n)
    error_result=circular_buffer_realloc(cb,cb->current_length+n,0) ;
  if(!error_result)
    {
      space=cb->buf_size-cb->current_index ;
      if(space>n)
        space=n ;
      memcpy(&(cb->buf[cb->current_index]),b,space) ;
      if(space<n)
        memcpy(cb->buf,&(b[space]),n-space) ;
      cb->current_index+=n ;
      for(;cb->current_index>=cb->buf_size;)
        cb->current_index-=cb->buf_size ;
      cb->current_length+=n ;
      if(cb->current_length>cb->buf_size)
        {
          /* this is fatal, but impossible to reach (will never happen) */
          fprintf(stderr,"circular_buffer_write: overflow! cur_length=%d, buf_size=%d\n",cb->current_length,cb->buf_size) ;
          cb->current_length=cb->buf_size ; 
       }
    }
  return(error_result) ;
}

int circular_buffer_prepend(struct circular_buffer_type *cb, char *b, int n)
{
  int error_result, cb_start, prefix_start, overflow ;

  error_result=0 ;
  if(cb->buf_size<cb->current_length+n)
    error_result=circular_buffer_realloc(cb,cb->current_length+n,0) ;
  if(!error_result)
    {
      cb_start=cb->current_index-cb->current_length ;
      for(;cb_start<0;cb_start+=cb->buf_size) ;
      prefix_start=cb_start-n ;
      for(;prefix_start<0;prefix_start+=cb->buf_size) ;
      if(prefix_start+n<=cb->buf_size)
        memcpy(&(cb->buf[prefix_start]),b,n) ;
      else
        {
          overflow=prefix_start+n-cb->buf_size ;
          memcpy(&(cb->buf[prefix_start]),b,n-overflow) ;
          memcpy(cb->buf,&(b[n-overflow]),overflow) ;
        }
      cb->current_length+=n ;
      if(cb->current_length>cb->buf_size)
        {
          /* this is fatal, but impossible to reach (will never happen) */
          fprintf(stderr,"circular_buffer_prepend: overflow! cur_length=%d, buf_size=%d\n",cb->current_length,cb->buf_size) ;
          cb->current_length=cb->buf_size ; 
       }
    }

  return(error_result) ;
}

int circular_buffer_peek(struct circular_buffer_type *cb, char *b, int n)
{
  int error_result, index, wrap ;

  error_result=0 ;
  if(cb->current_length<n)
    error_result= -2 ;
  if(!error_result)
    {
      index=n+cb->current_index-cb->current_length ;
      if(index<0)
        index+=cb->buf_size ;
      wrap=n-index ;

      if(wrap>0)
        {
          memcpy(b,&(cb->buf[cb->buf_size-wrap]),wrap) ;
          memcpy(&(b[wrap]),cb->buf,index) ;
        }
      else
        memcpy(b,&(cb->buf[index-n]),n) ;
    }

  return(error_result) ;
}


int circular_buffer_read(struct circular_buffer_type *cb, char *b, int n)
{
  int error_result ;

  error_result=circular_buffer_peek(cb,b,n) ;
  if(!error_result)
    cb->current_length-=n ;

  return(error_result) ;
}
